;;;; Copyright © 2016 Federico Beffa <beffa@fbengineering.ch>
;;;;
;;;; This library is free software; you can redistribute it and/or
;;;; modify it under the terms of the GNU Lesser General Public
;;;; License as published by the Free Software Foundation; either
;;;; version 3 of the License, or (at your option) any later version.
;;;; 
;;;; This library is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;;; Lesser General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU Lesser General Public
;;;; License along with this library; if not, write to the Free Software
;;;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
;;;; 

;;;; This code is based on the restart SRFI draft written by Taylor
;;;; Campbell:
;;;;
;;;; http://mumble.net/~campbell/proposals/restart.scm
;;;;
;;;; The original copyright reads:

;;;; This code is written by Taylor Campbell and placed in the Public
;;;; Domain.  All warranties are disclaimed.

;;; Commentary:

;;; This library provides the means to restart a computation after an
;;; exception is raised, similar to Common Lisp and MIT/GNU Scheme
;;; restarts.

;;; Code

(define-module (restart restart-rnrs)
  #:use-module ((srfi srfi-1))
  ;; #:use-module ((ice-9 control))
  #:use-module ((rnrs records syntactic))
  #:use-module ((rnrs exceptions))
  #:use-module ((rnrs conditions))
  #:export (restart
            restart-interactively
            current-restarters
            with-restarter
            call-with-restarter
            call-with-interactive-restarter
            with-exiting-restarter
            call-with-exiting-restarter
            find-restarter

            make-restarter
            restarter?
            restarter-tag
            restarter-description
            restarter-invoker
            restarter-interactor
            restarter-list?
            
            restartable-condition?
            condition-restarters
            make-restartable-condition
            restartable-condition
            error-with-current-restarters
            restart-case
            interactive-restarters?
            abort-restarter
            make-restart-invoker
            use-value
            abort))

;; Using SRFI-9 records
;; (define-record-type <restarter>
;;   (make-restarter tag description invoker interactor)
;;   restarter?
;;   (tag restarter-tag)
;;   (description restarter-description)
;;   (invoker restarter-invoker)
;;   (interactor restarter-interactor))

;; Using (rnrs) records
(define-record-type (<resterter> make-restarter restarter?)
  (fields (immutable tag restarter-tag)
          (immutable description restarter-description)
          (immutable invoker restarter-invoker)
          (immutable interactor restarter-interactor)))

(define (restart spec . args)
  (let ((win (lambda (r) (apply (restarter-invoker r) args))))
    (cond ((restarter? spec) (win spec))
          ((find-restarter spec) => win)
          (else (error "invalid restarter specifier argument"
                       `(RESTART ,spec ,@args))))))

(define (restart-interactively spec)
  (let ((win (lambda (r)
               (call-with-values (restarter-interactor r)
                                 (restarter-invoker r)))))
    (cond ((restarter? spec) (win spec))
          ((find-restarter spec) => win)
          (else (error "invalid restarter specifier argument"
                       `(RESTART-INTERACTIVELY ,spec))))))

(define current-restarters (make-parameter '()))

(define (restarter-list? list)
  (and (list? list) (every restarter? list)))

(define (with-restarter restarter thunk)
  (parameterize ((current-restarters
                  (cond
                   ((restarter-list? restarter)
                    (append restarter (current-restarters)))
                   ((restarter? restarter)
                    (cons restarter (current-restarters)))
                   (else
                    (error with-restarter "invalid restarter argument"
                           restarter)))))
    (thunk)))

(define (call-with-restarter tag desc invoker receiver)
  (call-with-interactive-restarter tag desc invoker #f receiver))

(define (call-with-interactive-restarter tag desc invoker interactor
          receiver)
  (let ((r (make-restarter tag desc invoker interactor)))
    (with-restarter r (lambda () (receiver r)))))

(define (with-exiting-restarter tag desc thunk)
  (call-with-exiting-restarter tag desc (lambda (r) (thunk))))

(define (call-with-exiting-restarter tag desc receiver)
  (call-with-current-continuation
    (lambda (k)
      (call-with-interactive-restarter tag desc
          (lambda () (k))               ; invoker
          (lambda () (values))          ; interactor
        receiver))))


;;; (rnrs conditions) extension

(define-condition-type &restartable &condition
  make-restartable restartable-condition?
  (restarters condition-restarters))

(define (make-restartable-condition restarters . components)
  (apply condition
         (make-restartable restarters)
         components))

(define (find-restarter tag . condition)
  (let loop ((restarters
              (if (or (null? condition) (not (car condition)))
                  (current-restarters)
                  (condition-restarters (car condition)))))
    (cond ((null? restarters) #f)
          ((eqv? (restarter-tag (car restarters)) tag)
           (car restarters))
          (else (loop (cdr restarters))))))

(define-syntax restarter-clause
  (syntax-rules ()
    ((_ (expression))
     expression)
    ((_ (tag desc invoker))
     (make-restarter tag desc invoker #f))
    ((_ (tag desc invoker interactor))
     (make-restarter tag desc invoker interactor))))

(define-syntax restartable-condition
  (syntax-rules ()
    ((_ (restarter ...)
       (condition-type field-value ...)
       ...)
     (make-restartable-condition
      (list (restarter-clause restarter) ...)
      (condition (condition-type field-value ...)
                 ...)))))

(define (error-with-current-restarters who msg . irritants)
  (raise (make-restartable-condition
          (current-restarters)
          (make-error)
          (make-message-condition msg)
          (make-who-condition who)
          (make-irritants-condition irritants))))


;;; restart-case Common Lisp stype, but not identical!

(define-syntax restart-case-clause
  (syntax-rules (without-continuation)
    ((_ (tag description invoker-formals invoker-body) k)
     (make-restarter 'tag description
                     (lambda invoker-formals (k invoker-body))
                     (lambda () (values))))
    ((_ (tag description invoker-formals without-continuation invoker-body) k)
     (make-restarter 'tag description
                     (lambda invoker-formals invoker-body)
                     (lambda () (values))))
    ((_ (tag description invoker-formals invoker-body interaktor) k)
     (make-restarter 'tag description
                     (lambda invoker-formals (k invoker-body))
                     interaktor))
    ((_ (tag description invoker-formals without-continuation invoker-body
             interaktor) k)
     (make-restarter 'tag description
                     (lambda invoker-formals invoker-body)
                     interaktor))))

;; (define-syntax restart-case
;;   (syntax-rules ()
;;     ((_ expression clause ...)
;;      (call-with-current-continuation
;;       ;; call-with-escape-continuation ; would be better ??
;;       (lambda (k)
;;         (with-restarter (list (restart-case-clause clause k) ...)
;;                         (lambda ()  expression)))))))

(define-syntax restart-case
  (syntax-rules ()
    ((_ expression clause ...)
     (call-with-current-continuation
      (lambda (k)
        (parameterize ((current-restarters
                        (list (restart-case-clause clause k) ...)))
          expression))))))


;;; Standard restarters helpers

(define interactive-restarters? (make-parameter #f))

(define abort-restarter
  (make-restarter 'abort "Abort computation."
                  (lambda () (throw 'quit))
                  (lambda () (values))))

(define-syntax make-restart-invoker
  (syntax-rules ()
    ((_ name argument ... not-found-thunk)
     (lambda* (argument ... #:optional (condition #f))
       (let ((restarter (find-restarter name condition)))
         (if restarter
             (if (interactive-restarters?)
                 (restart-interactively restarter)
                 (restart restarter argument ...))
             (not-found-thunk)))))))

(define use-value (make-restart-invoker
                   'use-value value (lambda () *unspecified*)))

(define abort (make-restart-invoker
               'abort (lambda () *unspecified*)))
